@@ -19,15 +19,15 @@ import datetime
 import warnings
 
 from bson import SON
-from bson.binary import (STANDARD, PYTHON_LEGACY,
-                         JAVA_LEGACY, CSHARP_LEGACY)
+from bson.binary import UuidRepresentation
 from bson.codec_options import CodecOptions, TypeRegistry
-from bson.py3compat import abc, integer_types, iteritems, string_type
+from bson.py3compat import abc, integer_types, iteritems, string_type, PY3
 from bson.raw_bson import RawBSONDocument
 from pymongo.auth import MECHANISMS
 from pymongo.compression_support import (validate_compressors,
                                          validate_zlib_compression_level)
 from pymongo.driver_info import DriverInfo
+from pymongo.server_api import ServerApi
 from pymongo.encryption_options import validate_auto_encryption_opts_or_none
 from pymongo.errors import ConfigurationError
 from pymongo.monitoring import _validate_event_listeners
@@ -43,6 +43,10 @@ try:
 except ImportError:
     ORDERED_TYPES = (SON,)
 
+if PY3:
+    from urllib.parse import unquote_plus
+else:
+    from urllib import unquote_plus
 
 # Defaults until we connect to a server and get updated limits.
 MAX_BSON_SIZE = 16 * (1024 ** 2)
@@ -54,9 +58,9 @@ MAX_WRITE_BATCH_SIZE = 1000
 # What this version of PyMongo supports.
 MIN_SUPPORTED_SERVER_VERSION = "2.6"
 MIN_SUPPORTED_WIRE_VERSION = 2
-MAX_SUPPORTED_WIRE_VERSION = 8
+MAX_SUPPORTED_WIRE_VERSION = 13
 
-# Frequency to call ismaster on servers, in seconds.
+# Frequency to call hello on servers, in seconds.
 HEARTBEAT_FREQUENCY = 10
 
 # Frequency to process kill-cursors, in seconds. See MongoClient.close_cursor.
@@ -71,7 +75,7 @@ EVENTS_QUEUE_FREQUENCY = 1
 # longest it is willing to wait for a new primary to be found.
 SERVER_SELECTION_TIMEOUT = 30
 
-# Spec requires at least 500ms between ismaster calls.
+# Spec requires at least 500ms between hello calls.
 MIN_HEARTBEAT_INTERVAL = 0.5
 
 # Spec requires at least 60s between SRV rescans.
@@ -128,13 +132,13 @@ def partition_node(node):
 
 
 def clean_node(node):
-    """Split and normalize a node name from an ismaster response."""
+    """Split and normalize a node name from a hello response."""
     host, port = partition_node(node)
 
     # Normalize hostname to lowercase, since DNS is case-insensitive:
     # http://tools.ietf.org/html/rfc4343
     # This prevents useless rediscovery if "foo.com" is in the seed list but
-    # "FOO.com" is in the ismaster response.
+    # "FOO.com" is in the hello response.
     return host.lower(), port
 
 
@@ -145,10 +149,11 @@ def raise_config_error(key, dummy):
 
 # Mapping of URI uuid representation options to valid subtypes.
 _UUID_REPRESENTATIONS = {
-    'standard': STANDARD,
-    'pythonLegacy': PYTHON_LEGACY,
-    'javaLegacy': JAVA_LEGACY,
-    'csharpLegacy': CSHARP_LEGACY
+    'unspecified': UuidRepresentation.UNSPECIFIED,
+    'standard': UuidRepresentation.STANDARD,
+    'pythonLegacy': UuidRepresentation.PYTHON_LEGACY,
+    'javaLegacy': UuidRepresentation.JAVA_LEGACY,
+    'csharpLegacy': UuidRepresentation.CSHARP_LEGACY
 }
 
 
@@ -329,6 +334,16 @@ def validate_timeout_or_zero(option, value):
     return validate_positive_float(option, value) / 1000.0
 
 
+def validate_timeout_or_none_or_zero(option, value):
+    """Validates a timeout specified in milliseconds returning
+    a value in floating point seconds. value=0 and value="0" are treated the
+    same as value=None which means unlimited timeout.
+    """
+    if value is None or value == 0 or value == "0":
+        return None
+    return validate_positive_float(option, value) / 1000.0
+
+
 def validate_max_staleness(option, value):
     """Validates maxStalenessSeconds according to the Max Staleness Spec."""
     if value == -1 or value == "-1":
@@ -391,8 +406,11 @@ def validate_read_preference_tags(name, value):
             tag_sets.append({})
             continue
         try:
-            tag_sets.append(dict([tag.split(":")
-                                  for tag in tag_set.split(",")]))
+            tags = {}
+            for tag in tag_set.split(","):
+                key, val = tag.split(":")
+                tags[unquote_plus(key)] = unquote_plus(val)
+            tag_sets.append(tags)
         except Exception:
             raise ValueError("%r not a valid "
                              "value for %s" % (tag_set, name))
@@ -401,7 +419,8 @@ def validate_read_preference_tags(name, value):
 
 _MECHANISM_PROPS = frozenset(['SERVICE_NAME',
                               'CANONICALIZE_HOST_NAME',
-                              'SERVICE_REALM'])
+                              'SERVICE_REALM',
+                              'AWS_SESSION_TOKEN'])
 
 
 def validate_auth_mechanism_properties(option, value):
@@ -412,6 +431,10 @@ def validate_auth_mechanism_properties(option, value):
         try:
             key, val = opt.split(':')
         except ValueError:
+            # Try not to leak the token.
+            if 'AWS_SESSION_TOKEN' in opt:
+                opt = ('AWS_SESSION_TOKEN:<redacted token>, did you forget '
+                       'to percent-escape the token with quote_plus?')
             raise ValueError("auth mechanism properties must be "
                              "key:value pairs like SERVICE_NAME:"
                              "mongodb, not %s." % (opt,))
@@ -422,7 +445,7 @@ def validate_auth_mechanism_properties(option, value):
         if key == 'CANONICALIZE_HOST_NAME':
             props[key] = validate_boolean_or_string(key, val)
         else:
-            props[key] = val
+            props[key] = unquote_plus(val)
 
     return props
 
@@ -503,6 +526,15 @@ def validate_driver_or_none(option, value):
     return value
 
 
+def validate_server_api_or_none(option, value):
+    """Validate the server_api keyword arg."""
+    if value is None:
+        return value
+    if not isinstance(value, ServerApi):
+        raise TypeError("%s must be an instance of ServerApi" % (option,))
+    return value
+
+
 def validate_is_callable_or_none(option, value):
     """Validates that 'value' is a callable."""
     if value is None:
@@ -581,7 +613,8 @@ URI_OPTIONS_VALIDATOR_MAP = {
     'authmechanismproperties': validate_auth_mechanism_properties,
     'authsource': validate_string,
     'compressors': validate_compressors,
-    'connecttimeoutms': validate_timeout_or_none,
+    'connecttimeoutms': validate_timeout_or_none_or_zero,
+    'directconnection': validate_boolean_or_string,
     'heartbeatfrequencyms': validate_timeout_or_none,
     'journal': validate_boolean_or_string,
     'localthresholdms': validate_positive_float_or_zero,
@@ -594,17 +627,21 @@ URI_OPTIONS_VALIDATOR_MAP = {
     'replicaset': validate_string_or_none,
     'retryreads': validate_boolean_or_string,
     'retrywrites': validate_boolean_or_string,
+    'loadbalanced': validate_boolean_or_string,
     'serverselectiontimeoutms': validate_timeout_or_zero,
-    'sockettimeoutms': validate_timeout_or_none,
+    'sockettimeoutms': validate_timeout_or_none_or_zero,
     'ssl_keyfile': validate_readable,
     'tls': validate_boolean_or_string,
     'tlsallowinvalidcertificates': validate_allow_invalid_certs,
     'ssl_cert_reqs': validate_cert_reqs,
+    # Normalized to ssl_match_hostname which is the logical inverse of tlsallowinvalidhostnames
     'tlsallowinvalidhostnames': lambda *x: not validate_boolean_or_string(*x),
     'ssl_match_hostname': validate_boolean_or_string,
     'tlscafile': validate_readable,
     'tlscertificatekeyfile': validate_readable,
     'tlscertificatekeyfilepassword': validate_string_or_none,
+    # Normalized to ssl_check_ocsp_endpoint which is the logical inverse of tlsdisableocspendpointcheck
+    'tlsdisableocspendpointcheck': lambda *x: not validate_boolean_or_string(*x),
     'tlsinsecure': validate_boolean_or_string,
     'w': validate_non_negative_int_or_basestring,
     'wtimeoutms': validate_non_negative_integer,
@@ -616,6 +653,7 @@ URI_OPTIONS_VALIDATOR_MAP = {
 NONSPEC_OPTIONS_VALIDATOR_MAP = {
     'connect': validate_boolean_or_string,
     'driver': validate_driver_or_none,
+    'server_api': validate_server_api_or_none,
     'fsync': validate_boolean_or_string,
     'minpoolsize': validate_non_negative_integer,
     'socketkeepalive': validate_boolean_or_string,
@@ -656,6 +694,7 @@ INTERNAL_URI_OPTION_NAME_MAP = {
     'tlscafile': 'ssl_ca_certs',
     'tlscertificatekeyfile': 'ssl_certfile',
     'tlscertificatekeyfilepassword': 'ssl_pem_passphrase',
+    'tlsdisableocspendpointcheck': 'ssl_check_ocsp_endpoint',
 }
 
 # Map from deprecated URI option names to a tuple indicating the method of
@@ -674,6 +713,14 @@ URI_OPTIONS_DEPRECATION_MAP = {
     'ssl_match_hostname': ('renamed', 'tlsAllowInvalidHostnames'),
     'ssl_crlfile': ('renamed', 'tlsCRLFile'),
     'ssl_ca_certs': ('renamed', 'tlsCAFile'),
+    'ssl_certfile': ('removed', (
+        'Instead of using ssl_certfile to specify the certificate file, '
+        'use tlsCertificateKeyFile to pass a single file containing both '
+        'the client certificate and the private key')),
+    'ssl_keyfile': ('removed', (
+        'Instead of using ssl_keyfile to specify the private keyfile, '
+        'use tlsCertificateKeyFile to pass a single file containing both '
+        'the client certificate and the private key')),
     'ssl_pem_passphrase': ('renamed', 'tlsCertificateKeyFilePassword'),
     'waitqueuemultiple': ('removed', (
         'Instead of using waitQueueMultiple to bound queuing, limit the size '
@@ -714,7 +761,7 @@ def validate_auth_option(option, value):
     if lower not in _AUTH_OPTIONS:
         raise ConfigurationError('Unknown '
                                  'authentication option: %s' % (option,))
-    return lower, value
+    return option, value
 
 
 def validate(option, value):
@@ -723,7 +770,7 @@ def validate(option, value):
     lower = option.lower()
     validator = VALIDATORS.get(lower, raise_config_error)
     value = validator(option, value)
-    return lower, value
+    return option, value
 
 
 def get_validated_options(options, warn=True):
@@ -932,4 +979,4 @@ class _CaseInsensitiveDictionary(abc.MutableMapping):
                 self[key] = other[key]
 
     def cased_key(self, key):
-        return self.__casedkeys[key.lower()]
\ No newline at end of file
+        return self.__casedkeys[key.lower()]
